/* main.c -- Single-profile interface for lidar multiple scattering
   algorithm

   Copyright (C) 2004-2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   Run "./multiscatter -help" to get usage information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "multiscatter.h"

#define MAX_GATES 1000
#define MAX_CHARS 128

#ifdef SINGLE_PRECISION
#define READ_HEADER_FORMAT "%d %g %g %g %g\n"
#define READ_FORMAT "%g %g %g %g %g\n"
#else
#define READ_HEADER_FORMAT "%d %lg %lg %lg %lg\n"
#define READ_FORMAT "%lg %lg %lg %lg %lg\n"
#endif

void
usage(char *exec_name)
{
  fprintf(stderr,
	  "Usage\n"
	  "  %s [options] < data_in.dat > data_out.dat\n"
	  "\n", exec_name);
  fprintf(stderr,
	  "Options\n"
	  "  -simple-optical-depth\n"
	  "                  Use simple optical depth integration\n"
	  "  -crude-optical-depth\n"
	  "                  Use crude optical depth integration\n"
	  "  -crude-integration\n"
	  "                  Use crude double/multiple scattering integration\n");
  fprintf(stderr,
	  "  -no-multiscat-within-gate\n"
	  "                  Photon cannot be forward scattered more than once within a single gate\n"
	  "                  Note that this has the opposite effect to \"-crude-integration\n"
	  "  -double-scattering-only\n"
	  "                  Don't include triple and higher-order scatterings\n");
  fprintf(stderr,
	  "  -no-molecular-extinction\n"
	  "                  Molecules do not extinguish the beam\n"
	  "  -wide-angle-cutoff <theta>\n"
	  "                  Forward scattering at angles greater than <theta> radians are deemed to\n"
	  "                  escape, a crude way to deal with a problem associated with aerosols\n");
  fprintf(stderr,
	  "  -crude-double-scattering\n"
	  "                  Don't use Eloranta's slow but accurate double scattering formulation\n"
	  "  -approx-exp     Appriximate the exp() function call for speed\n"
	  "  -jacobian       Output the Jacobian instead\n"
	  "  -numerical-jacobian\n"
	  "                  Output the Jacobian calculated (slowly) using finite differences\n"
);
  
  fprintf(stderr,
	  "  -output-bscats  Also output 6 individual backscatter components\n"
	  "  -output-all     Also output 8 internal variables at each gate\n"
	  "  -repeat n       Repeat algorithm n times (for benchmarking)\n"
	  "  -help           Display this message\n"
	  "\n");
  fprintf(stderr,
	  "Input data\n"
	  "  First line: 5 values\n"
	  "    1: Number of range gates in profile\n"
	  "    2: Lidar wavelength (microns)\n"
	  "    3: Laser divergence, 1/e half-width (radians)\n"
	  "    4: Telescope field-of-view, half-width (radians)\n"
	  "    5: Altitude of lidar (m)\n");
  fprintf(stderr,
	  "  Subsequent lines: 4 or 5 values\n"
	  "    1: Range of gate above ground starting with nearest gate to lidar (m)\n"
	  "    2: Extinction coefficient of cloud/aerosol only (m-1)\n"
	  "    3: Equivalent-area cloud/aerosol radius (microns)\n"
	  "    4: Extinction-to-backscatter ratio of cloud/aerosol (sr)\n"
	  "    5: Air extinction coefficient (m-1); optional (zero if missing)\n"
	  "\n");
  fprintf(stderr,
	  "Output data (without -jacobian or -numerical-jacobian option)\n"
	  "  Default: 5 values per range gate\n"
	  "    1: Index of range gate\n"
	  "    2: Range above ground (m)\n"
	  "    3: Extinction coefficient (m-1)\n"
	  "    4: Equivalent-area particle radius (microns)\n"
	  "    5: Attenuated backscatter, total (m-1 sr-1)\n");
  fprintf(stderr,
	  "  With the \"-output-bscats\" option, output an additional 6 values\n"
	  "    6: Atten. backscatter, single scattering cloud/aerosol only (m-1 sr-1)\n"
	  "    7: Atten. backscatter, double scattering cloud/aerosol only (m-1 sr-1)\n"
	  "    8: Atten. backscatter, triple+ scattering cloud/aerosol (m-1 sr-1)\n"
	  "    9: Atten. backscatter, single scattering air only (m-1 sr-1)\n"
	  "   10: Atten. backscatter, double scattering air only (m-1 sr-1)\n"
	  "   11: Atten. backscatter, triple+ scattering air only (m-1 sr-1)\n");
  fprintf(stderr,
	  "  With the \"-output-all\" option, output an additional 8 values\n"
	  "   12: Es, singly forward scattered energy coefficient\n"
	  "   13: Em, multiply forward scattered energy coefficient\n"
	  "   14: lateral standard deviation of outgoing singly forward scattered photons (m)\n"
	  "   15: lateral standard deviation of outgoing multply forward scattered photons (m)\n");
  fprintf(stderr,
	  "   16: standard deviation of singly forward scattered photon direction (radians)\n"
	  "   17: standard deviation of multiply forward scattered photon direction (radians)\n"
	  "   18: correlation coefficient of singly scattered photon angular position and direction\n"
	  "   19: correlation coefficient of multiply scattered photon angular position and direction\n"
	  "\n");
  fprintf(stderr,
	  "Output data (with -jacobian or -numerical-jacobian option)\n"
	  "   n x n matrix: d(ln atten. backscatter) / d(ln extinction coefficient)\n"
	  "   n x n matrix: d(ln atten. backscatter) / d(ln particle radius)\n"
	  "   1 x n vector: d(ln atten. backscatter) / d(ln ext-to-bscat ratio)\n");
}


int
main(int argc, char **argv)
{
  int n, i, iarg;
  ms_real wavelength, rho_laser, rho_telescope, lidar_altitude;
  int nrepeats = 1;
  int output_bscats = 0;
  int output_all = 0;
  int output_jacobian = 0;
  int ninputs;
  char buffer[MAX_CHARS];
  int options = 0;
  int ch;

  /* Ought to dynamically allocate these */
  ms_real range[MAX_GATES], radius[MAX_GATES],
    ext[MAX_GATES], ext_bscat_ratio[MAX_GATES], ext_air[MAX_GATES],
    bscat[MAX_GATES];

  /* Handle command-line arguments */
  for (iarg = 1; iarg < argc; ++iarg) {
    if (strcmp(argv[iarg], "-output-bscats") == 0) {
      /* Output different backscatters */
      output_bscats = 1;
    }
    else if (strcmp(argv[iarg], "-output-all") == 0) {
      /* Output all internal variables */
      output_bscats = output_all = 1;
    }
    else if (strcmp(argv[iarg], "-simple-optical-depth") == 0) {
      /* Use simple optical depth calculation */
      options |= MS_SIMPLE_OPTICAL_DEPTH;
    }
    else if (strcmp(argv[iarg], "-crude-optical-depth") == 0) {
      /* Use crude optical depth calculation */
      options |= MS_CRUDE_OPTICAL_DEPTH;
    }
    else if (strcmp(argv[iarg], "-crude-integration") == 0) {
      /* Use crude integration for double and multiple scattering */
      options |= MS_CRUDE_INTEGRATION;
      options &= ~MS_NO_MULTISCAT_WITHIN_GATE;
    }
    else if (strcmp(argv[iarg], "-no-multiscat-within-gate") == 0) {
      /* Each photon cannot be forward scattered more than once within
	 a range gate */
      options |= MS_NO_MULTISCAT_WITHIN_GATE;
      options &= ~MS_CRUDE_INTEGRATION;
    }
    else if (strcmp(argv[iarg], "-no-molecular-extinction") == 0) {
      /* Molecules backscatter but don't extinguish */
      options |= MS_NO_MOLECULAR_EXTINCTION;
    }
    else if (strcmp(argv[iarg], "-crude-double-scattering") == 0) {
      /* Use crude but fast method for double scattering */
      options |= MS_CRUDE_DOUBLE_SCATTERING;
    }
    else if (strcmp(argv[iarg], "-approx-exp") == 0) {
      /* Use crude but very fast method for double scattering */
      options |= MS_APPROXIMATE_EXPONENTIAL;
    }
    else if (strcmp(argv[iarg], "-double-scattering-only") == 0) {
      /* Use crude but fast method for double scattering */
      options |= MS_DOUBLE_SCATTERING_ONLY;
    }
    else if (strcmp(argv[iarg], "-wide-angle-cutoff") == 0) {
      /* A crude method to cope with the problem of representing both
	 cloud and wide-angle-scattering aerosol in the same
	 profile */
      if (iarg + 1 < argc) {
	options |= MS_WIDE_ANGLE_CUTOFF;
	ms_max_theta = atof(argv[++iarg]);
      }
      else {
	fprintf(stderr, "Option \"-wide-angle-cutoff\" requires an angle to be specified\n");
	exit(2);
      }
    }
    else if (strcmp(argv[iarg], "-jacobian") == 0) {
      /* Output the Jacobian */
      output_jacobian = 1;
    }
    else if (strcmp(argv[iarg], "-numerical-jacobian") == 0) {
      /* Output the Jacobian */
      output_jacobian = 1;
      options |= MS_NUMERICAL_JACOBIAN;
    }
    else if (strcmp(argv[iarg], "-repeat") == 0) {
      /* Perform repeat calculations for benchmarking purposes */
      if (argc > ++iarg) {
	nrepeats = atoi(argv[iarg]);
	if (nrepeats < 1 || nrepeats > 1000000) {
	  fprintf(stderr,
		  "Error: number of repeats must be between 1 and 1000000\n");
	  exit(1);
	}
      }
      else {
	fprintf(stderr, "Error: number of repeats not specified\n");
	exit(1);
      }
    }
    else {
      /* Argument not understood */
      usage(argv[0]);
      exit(0);
    }
  }

  /* Set options */
  ms_set_options(options);

  /* Skip commented lines */
  while ((ch = fgetc(stdin)) == '#') {
    while ((ch = fgetc(stdin)) != '\n') {
      if (ch == EOF) {
	fprintf(stderr, "Only comments found on standard input\n");
	exit(2);
      }
    }
  }
  ungetc(ch, stdin);

  /* Read input data */
  ninputs = fscanf(stdin, READ_HEADER_FORMAT, &n, &wavelength,
		   &rho_laser, &rho_telescope, &lidar_altitude);
  if (ninputs < 5) {
    fprintf(stderr, "Garbled input at first line: "
	    "only %d inputs read successfully\n", ninputs);
    exit(2);
  }
  if (n > MAX_GATES) {
    fprintf(stderr, "Too many range gates: maximum %d\n", MAX_GATES);
  }
  for (i = 0; i < n; i++) {
    fgets(buffer, MAX_CHARS, stdin);
    ninputs = sscanf(buffer, READ_FORMAT, range+i, ext+i,
		     radius+i, ext_bscat_ratio+i, ext_air+i);
    if (ninputs < 4) {
      fprintf(stderr, "Garbled input at line %d: "
	      "only %d inputs read successfully %f %f\n", 
	      i+1, ninputs, range[i], range[i-1]);
      exit(2);
    }
    else if (ninputs < 5) {
      ext_air[i] = 0.0;
    }
  }

  if (!output_jacobian) {
  /* Call multiscatter algorithm */
  for (i = 0; i < nrepeats; i++) {
    multiscatter(n, wavelength, rho_laser, rho_telescope,
		 lidar_altitude, range, radius,
		 ext, ext_bscat_ratio, ext_air,
		 bscat);
  }
  /* Send output to stdout */
  for (i = 0; i < n; i++) {
    fprintf(stdout, "%d %g %g %g %g",
	    i+1, 
	    (float) range[i],
	    (float) ext[i],
	    (float) radius[i],
	    (float) bscat[i]);
    if (output_bscats) {
      fprintf(stdout, " %g %g %g %g %g %g",
	      (float) ms_bscat_single[i],
	      (float) ms_bscat_double[i],
	      (float) ms_bscat_multi[i],
	      (float) ms_bscat_air_single[i],
	      (float) ms_bscat_air_double[i],
	      (float) ms_bscat_air_multi[i]);
      if (output_all) {
	fprintf(stdout, " %g %g %g %g %g %g %g %g",
		(float) ms_E_once[i],
		(float) ms_E_multi[i],
		(float) sqrt(ms_Emu2_once[i]/ms_E_once[i]),
		(float) sqrt(ms_Emu2_multi[i]/ms_E_multi[i]),
		(float) sqrt(ms_Pmu2_once[i]/ms_E_once[i]), 
		(float) sqrt(ms_Pmu2_multi[i]/ms_E_multi[i]),
		(float) ms_EPcov_once[i]/sqrt(ms_Emu2_once[i]*ms_Pmu2_once[i]), 
		(float) ms_EPcov_multi[i]/sqrt(ms_Emu2_multi[i]*ms_Pmu2_multi[i]));
      }
    }
    fprintf(stdout, "\n");
  }
  }
  else {
    /* Calculate the Jacobian */
    ms_real *jacobian_data = malloc((2*n+1)*n*sizeof(ms_real));
    ms_real **d_ln_beta_d_ln_ext = malloc(n*sizeof(ms_real *));
    ms_real **d_ln_beta_d_ln_radius = malloc(n*sizeof(ms_real *));
    ms_real *d_ln_beta_d_ln_ext_bscat_ratio = jacobian_data + 2*n*n;
    int *i_xy = malloc(n*sizeof(int));
    int i, j;
    for (i = 0; i < n; i++) {
      d_ln_beta_d_ln_ext[i] = jacobian_data + i*n;
      d_ln_beta_d_ln_radius[i] = jacobian_data + (n+i)*n;
      i_xy[i] = i;
    }
    for (i = 0; i < nrepeats; i++) {
      multiscatter_with_jacobian(
		 n, wavelength, rho_laser, rho_telescope, lidar_altitude, 
		 range, radius, ext, ext_bscat_ratio, ext_air,
		 n, n, i_xy, i_xy, 
		 bscat,
		 d_ln_beta_d_ln_ext, d_ln_beta_d_ln_radius, 
		 d_ln_beta_d_ln_ext_bscat_ratio);
    }
    for (i = 0; i < n; i++) {
      for (j = 0; j < n; j++) {
	fprintf(stdout, " %g", d_ln_beta_d_ln_ext[i][j]);
      }
      fprintf(stdout, "\n");
    }
    for (i = 0; i < n; i++) {
      for (j = 0; j < n; j++) {
	fprintf(stdout, " %g", d_ln_beta_d_ln_radius[i][j]);
      }
      fprintf(stdout, "\n");
    }
    for (i = 0; i < n; i++) {
      fprintf(stdout, " %g", d_ln_beta_d_ln_ext_bscat_ratio[i]);
    }
    fprintf(stdout, "\n");
  }

  exit(0);
}
