/* multiscatter.h -- Calculation of lidar multiple scattering 

   Copyright (C) 2004-2005 Robin Hogan <r.j.hogan@reading.ac.uk>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


   This is the header file for an algorithm for efficient calculation
   of the lidar backscatter profile in the presence of multiple
   scattering.  It uses Eloranta's formulation for double scattering
   with a multi-moment parameterisation of the photon distribution for
   estimating scattering at higher orders that ensures O(N^2)
   efficiency where N is the number of range gates. 
*/

#ifndef _MULTISCATTER_H
#define _MULTISCATTER_H 1

#ifdef __cplusplus
extern "C" {
#endif

/* The code may be compiled either in single or double
   precision. Double precision is the default but may be overridden by
   compiling with "-DSINGLE_PRECISION". */
#ifdef SINGLE_PRECISION
typedef float ms_real;
#else
typedef double ms_real;
#endif

/* External functions return an integer with one of these values */
#define MS_SUCCESS 0
#define MS_FAILURE 1

/* Bitwise options */
#define MS_INTERMEDIATE_FOV (1L<<0) /* Deprecated */
#define MS_WIDE_FOV (1L<<1)         /* Deprecated */
#define MS_SIMPLE_OPTICAL_DEPTH (1L<<2)
#define MS_CRUDE_OPTICAL_DEPTH (1L<<3)
#define MS_CRUDE_INTEGRATION (1L<<4)
#define MS_NO_MULTISCAT_WITHIN_GATE (1L<<5)
#define MS_DOUBLE_SCATTERING_ONLY (1L<<6)
#define MS_CRUDE_DOUBLE_SCATTERING (1L<<7)
#define MS_NUMERICAL_JACOBIAN (1L<<8)
#define MS_NO_MOLECULAR_EXTINCTION (1L<<9)
#define MS_WIDE_ANGLE_CUTOFF (1L<<10)
#define MS_APPROXIMATE_EXPONENTIAL (1L<<11)

/* After calling multiscatter(), the following variables are available
   to be accessed - see multiscatter.c for details. */
extern ms_real *ms_E_once;           /* dimensionless */
extern ms_real *ms_Emu2_once;        /* radians^2 */
extern ms_real *ms_E_multi;          /* dimensionless */
extern ms_real *ms_Emu2_multi;       /* radians^2 */
extern ms_real *ms_Pmu2_once;        /* radians^2 */
extern ms_real *ms_Pmu2_multi;       /* radians^2 */
extern ms_real *ms_EPcov_once;       /* radians^2 */
extern ms_real *ms_EPcov_multi;      /* radians^2 */
extern ms_real *ms_bscat_single;     /* m-1 sr-1 */
extern ms_real *ms_bscat_double;     /* m-1 sr-1 */
extern ms_real *ms_bscat_multi;      /* m-1 sr-1 */
extern ms_real *ms_bscat_air_single; /* m-1 sr-1 */
extern ms_real *ms_bscat_air_double; /* m-1 sr-1 */
extern ms_real *ms_bscat_air_multi;  /* m-1 sr-1 */

extern int ms_options;
extern ms_real ms_max_theta;

/* Functions to allocate and zero or to free intermediate
   arrays. External programs need not call these functions.*/
int ms_init_intermdiate_arrays(int ngates);
void ms_free_intermediate_arrays();

/* Functions to change the algorithm options */
void ms_set_options(int options);
void ms_add_options(int options);

/* The multiple scattering algorithm is called using this
   function. MS_FAILURE is returned if there was a problem allocating
   the memory for the intermediate arrays, MS_SUCCESS otherwise. */
int multiscatter(
    /* Input data */
    int n,                    /* Number of range gates in profile */
    ms_real wavelength,       /* Wavelength of laser, microns */
    ms_real rho_laser,        /* 1/e laser beam divergence, radians */
    ms_real rho_telescope,    /* Half-angle field of view, radians */
    ms_real lidar_alt,        /* Lidar altitude, metres */
    ms_real *range,           /* Height of each range gate, metres */
    ms_real *radius,          /* Cloud/aerosol equivalent radius, microns */
    ms_real *ext,             /* Cloud/aerosol extinction coefficient, m-1 */
    ms_real *ext_bscat_ratio, /* Cloud/aerosol ext./backscatter ratio, sr */
    ms_real *ext_air,         /* Air ext. coefft, m-1 (NULL for vacuum) */
    /* Output data */
    ms_real *bscat_out        /* Measured backscatter, m-1 sr-1 */
    );

int multiscatter_with_jacobian(
    /* Input data */
    int n,                    /* Number of range gates in profile */
    ms_real wavelength,       /* Wavelength of laser, microns */
    ms_real rho_laser,        /* 1/e laser beam divergence, radians */
    ms_real rho_telescope,    /* Half-angle field of view, radians */
    ms_real lidar_alt,        /* Lidar altitude, metres */
    ms_real *range,           /* Height of each range gate, metres */
    ms_real *radius,          /* Cloud/aerosol equivalent radius, microns */
    ms_real *ext,             /* Cloud/aerosol extinction coefficient, m-1 */
    ms_real *ext_bscat_ratio, /* Cloud/aerosol ext./backscatter ratio, sr */
    ms_real *ext_air,         /* Air ext. coefft, m-1 (NULL for vacuum) */
    int n_x, /* Number of state variables to compute the Jacobian of */
    int n_y, /* Number of measurements to compute the Jacobian of */
    int *i_x, /* Index to the required state variables */
    int *i_y, /* Index to the required measurements */
    /* Output data */
    ms_real *bscat_out,        /* Measured backscatter, m-1 sr-1 */
    ms_real **d_ln_bscat_d_ln_ext, /* Jacobian with respect to ln(extinction) */
    ms_real **d_ln_bscat_d_ln_radius, /* Jacobian w.r.t. ln(radius) */
    ms_real *d_ln_bscat_d_ln_ext_bscat_ratio /* ...w.r.t. ln(ext_bscat_ratio) */
    );

#ifdef __cplusplus
}
#endif

#endif
